#! /usr/bin/python

## g2p.py
## Copyright (C) 2007 Ivan Uemlianin
## Contact: ivan@llaisdy.com
## 
## This file is part of Trefnydd.
##
## Trefnydd is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## Trefnydd is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with Trefnydd (i.e., doc/gpl.txt); if not, write to the Free Software
## Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
## or see http://www.gnu.org/licenses/gpl.txt


"""
========
trefnydd
========

Package:   trefnydd
Module:    g2p
Copyright: 2007 Ivan A. Uemlianin
Contact:   ivan@llaisdy.com
Licence:   GNU GPL (see doc/gpl.txt)

g2p
===

"""


import os
import copy, pprint
from trefnydd.utils import file_IO, location
from trefnydd.lib import phones

# TODO: how to implement priority? method or attribute?
#       in g2pruleset and gset
#
# 100107: basic g2p working!!!
# next: pd & pdEntry

class G2PRule(object):
    def __init__(self):
        self.graphemeSet = None
        self.phoneSet = None
        self.antecedent = None # GPString
        self.consequent = None # GPString
        self._name = ''
        self.comment = ''

    def getName(self):
        if self._name:
            return self._name
        elif self.antecedent or self.consequent:
            return self.__str__()
        else:
            return '# unnamed rule'

    def setName(self, name):
        self._name = name

    name = property(fset=setName, fget=getName)
    

    def __str__(self):
        out = u'%s  -->  %s' % (self.antecedent, self.consequent)#.encode('utf-8')
        return out

    def __repr__(self):
        out = self.__str__()
        if self.name or self.comment:
            out = '%s   # %s: %s' % (out, self.name, self.comment)
        return out

    def setGraphemeSet(self, gset):
        self.graphemeSet = gset

    def setPhoneSet(self, pset):
        self.phoneSet = pset

    def load(self, line):
        # line should be g(s) :p(s) # name: comment
        # line format
        # example: ll a n  -->  lh ae n   # name: comment
        name = comment = ''
        if '#' in line:
            meta = line[line.index('#'):].strip()
            line = line[:line.index('#')].strip()
            if ':' in meta:
                name, meta = meta.split(':')
                self.name = name.strip()
            self.comment = meta.strip()
        gstr, plist = line.split('-->')

        gs = GraphemeString([self.graphemeSet.getGraphemeByString(g)
                             for g in gstr.split()])
        gs.setGraphemeSet(self.graphemeSet)
        self.antecedent = gs
        ps = phones.PhoneString()
        ps.setPhoneset(self.phoneSet)
        ps.setPhoneString(' '.join(plist.split()))
        self.consequent = ps

    def transform(self, gpString, history=False):
        """
        Replace instances in gpString of antecedent with consequent
        return new gpString
        """
        new_gpString = gpString.replace(self.antecedent, self.consequent)
        if history is True:
            fired = False
            if self.antecedent in gpString:
                fired = True
            return new_gpString, fired
        else:
            return new_gpString

        
class G2PRuleSet(set, location.DataSetContainer):
    def __init__(self, fn=''):
        location.DataSetContainer.__init__(self)
        self.ext = '.g2p'
        self.graphemeSet = None
        self.phoneSet = None
        if fn:
            self.loadG2PRuleSet(fn)

    def getPreamble(self):
        return """# trefnydd::G2PRuleSet\n
graphemeSet =  %s
phoneSet =     %s
priorityFunc = %s
comment =      %s\n\n""" \
        % (self.graphemeSet.name, self.phoneSet.name,
           self.priorityFunc, self.comment)

    preamble = property(getPreamble)

    def loadNameValuePair(self, name, value, loadGPSets=False):
        if loadGPSets:
            if name.lower() == 'graphemeset':
                gset = GraphemeSet(value) # value = fn
                self.setGraphemeSet(gset)
            elif name.lower() == 'phoneset':
                pset = phones.Phoneset()
                pset.loadPhoneset(value)
                self.setPhoneSet(pset)
            else:
                setattr(self, name, value)
        else:
            setattr(self, 'src_fn_setting_%s' % name, value)
        # todo: set properly to objs not just fns

    def loadDataItem(self, line):
        rule = G2PRule()
        rule.setGraphemeSet(self.graphemeSet)
        rule.setPhoneSet(self.phoneSet)
        rule.load(line)
        self.add(rule)

    def addRule(self, rule, priority=None):
        """
        default priority = int( 100 / len(rule) )?
        """
        pass

    def removeRule(self, rule):
        pass

    def getRulesByPriority(self):
        #todo: yield?
        if not self.priorityFunc:
            # return list of rules by len(ant)
            l = [(r.antecedent.weight(), r) for r in self]
            l.sort()
            l.reverse()
            l = [r[1] for r in l]
            return l
        else:
            pass

    def setGraphemeSet(self, gset):
        self.graphemeSet = gset

    def setPhoneSet(self, pset):
        self.phoneSet = pset

    def transform(self, gpString, rules=False):
        if not rules:
            for rule in self.getRulesByPriority():
                gpString = rule.transform(gpString)
            return gpString
        else:
            rulesFired = []
            for rule in self.getRulesByPriority():
                gpString, fired = rule.transform(gpString, rules)
                if fired is True:
                    rulesFired.append(rule)
            return gpString, rulesFired


class Grapheme(object):
    def __init__(self, s=u''):
        self.string = u''
        self.comment = u''
        if s:
            self.string = s

    def getName(self):
        return self.string

    name = property(getName)

    def __str__(self):
        return '|%s|' % self.string

    def __repr__(self):
        out = '%s' % self.string
        if self.comment:
            out = '%s\t#  %s' % (out, self.comment)
        return out

    def __ne__(self, other):
        # todo: what about being from different graphemeSets!?
        #return self.string != other.string
        return self.string != other.__dict__.get('string')

    def loadLine(self, line):
        if '#' in line:
            line, comment = line.split('#')
            self.comment = comment.strip()
        self.string = line.strip()

### Python problem:
### what about Grapheme == Phone!?
### C++: bool operator_eq(Grapheme, Phone)


class GraphemeSet(set, location.DataSetContainer):
    # includes whole set as well as intermediates like V (= a e i o u etc)
    def __init__(self, fn=''):
        location.DataSetContainer.__init__(self)
        self.ext = '.graphemeset'
        if fn:
            self.loadGraphemeSet(fn)

    def getPreamble(self):
        return """# trefnydd::GraphemeSet\n
priorityFunc = %s
comment =      %s\n\n""" % (self.priorityFunc, self.comment)

    preamble = property(getPreamble)

    def loadNameValuePair(self, name, value):
        if name == 'priorityFunc':
            self.priorityFunc = value
        elif name.lower() == 'comment':
            self.comment = value
        else:
            setattr(self, name, value)

    def loadDataItem(self, line):
        g = Grapheme()
        g.loadLine(line.strip())
        self.addGrapheme(g)

    def addGrapheme(self, grapheme, priority=None):
        # default priority = int( 100 / len(self.string) )
        # longer graphemes, higher priority
        self.add(grapheme)

    def removeGrapheme(self, grapheme):
        pass

    def getGraphemesByPriority(self):
        if not self.priorityFunc: # order by length
            gl = [(len(g.string), g) for g in self]
            gl.sort()
            gl.reverse()
            return [g[1] for g in gl]
        else:
            # todo
            pass

    def getGraphemeByString(self, str):
        for g in self:
            if g.string == str:
                return g
        return False
            

class GPString(list):
    """
    string of mixed graphemes and phonemes
    """
    
    def __init__(self, gpList=[]):
        list.__init__(self, gpList)
        self.graphemeSet = None
        self.phoneSet = None

    def __eq__(self, other):
        for i in zip(self, other):
            #print 'GPString.__eq__():', i[0], i[1]
            if i[0] != i[1]:
                return False
        return True
            
    def __ne__(self, other):
        for i in zip(self, other):
            #print 'GPString.__eq__():', i[0], i[1]
            if i[0] != i[1]:
                return True
        return False
            
    def __contains__(self, other):
        leno = len(other)
        testlist = [self[i:i+leno]
                    for i in range(len(self)-leno+1)]
        for t in testlist:
            if other == t:
                return True
        return False

    def find(self, other):
        leno = len(other)
        testlist = [self[i:i+leno]
                    for i in range(len(self)-leno+1)]
        for t in range(len(testlist)):
            if other == testlist[t]:
                return t
        return -1
    
    def setGraphemeSet(self, gset):
        self.graphemeSet = gset

    def setPhoneSet(self, pset):
        self.phoneSet = pset

    def weight(self):
        a = sum([len(i.string) for i in self])/len(self)
        return eval('%s.%s' % (len(self), a))

    def replace(self, before, after):
        """
        Just like the string method
        """
        new = copy.copy(self) # or deepcopy?
        while before in new:
            idx = new.find(before)
            #print 'GPString.replace/while'
            pre = new[:idx]
            #print 'ok'
            post = new[idx+len(before):]
            new[:] = pre
            new.extend(after)
            new.extend(post)
        return new

# todo: should PhoneString be subclass of GPString?


class GraphemeString(GPString):
    def __init__(self, gList=[]):
        GPString.__init__(self, gList)
        self.graphemeSet = None

    def __str__(self):
        return '[%s]' % ' '.join(i.__str__() for i in self)

    def setGraphemeSet(self, gset):
        self.graphemeSet = gset

    def importString(self, s):
        """
        convert orthographic string to graphemes
        """
        glist = self.graphemeSet.getGraphemesByPriority()
        print '*** importString: %s' % s
        gs = self.string2gstring(s, glist)
        self[:] = gs

    def string2gstring(self, s, graphemeList, indent=0):
        tab = '  '*indent
        gs = []
        print '*** string2gstring'
        print graphemeList
        ng = graphemeList.pop(0)
        ngs = ng.string
        print 'ngs: %s' % ngs
        if ngs in s:
            gs = self.splitStringByGrapheme(s, ng)
            #print '%s[%s] gs: %s' % (tab, ngs, gs)
            i = 0
            while i < len(gs):
                if gs[i] not in self.graphemeSet:
                    x = self.string2gstring(gs[i],
                                            copy.copy(graphemeList),
                                            #graphemeList,
                                            indent+1)
                    tl = gs[:i]
                    tl.extend(x)
                    tl.extend(gs[i+1:])
                    gs = tl
                i += 1
        else:
            #gs = self.string2gstring(s, copy.copy(graphemeList), indent)
            gs = self.string2gstring(s, graphemeList, indent)
        return gs

                
    def splitStringByGrapheme(self, s, g):
        if s == g.string:
            return [g]
        sl = s.split(g.string)
        nsl = []
        if sl[0] == '':
            nsl.append(g)
            sl.pop(0)
        for i in sl:
            nsl.append(i)
            nsl.append(g)
        if sl[-1] == '':
            nsl.pop()
        nsl.pop()
        return nsl

## GRAPHEME_SET_CYMRAEG = GraphemeSet()
## GRAPHEME_SET_CYMRAEG.loadGraphemeSet('../meta/cymraeg.graphemeset')

## G2PRULESET_CYMRAEG = G2PRuleSet()
## G2PRULESET_CYMRAEG.loadG2PRuleSet('../meta/cymraeg.g2p')


def main():
    #gair = 'angharad' # ok
    #gair = 'llyfrgell' # ok
    #gair = 'llanfairpwllgwyngyll' # ok
    #gair = 'llyfrgellydd' # ok
    #gair = 'llefrith' # ok
    #gair = 'llanddeiniolen'
    gair = 'llangollen'  # should go to [lh ae n g o lh eh n]
    gs = GraphemeString()
    gs.setGraphemeSet(GRAPHEME_SET_CYMRAEG)
    gs.importString(gair)
    print 'gString: ', gs

    ps, rules = G2PRULESET_CYMRAEG.transform(gs, True)
    print 'pString: ', ps
    print '\n'.join(['%s' % r  for r in rules])

if __name__ == '__main__':
    main()

