#! /usr/bin/python

## audioFileInterface.py
## Copyright (C) 2007 Ivan Uemlianin
## Contact: ivan@llaisdy.com
## 
## This file is part of Trefnydd.
##
## Trefnydd is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## Trefnydd is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with Trefnydd (i.e., doc/gpl.txt); if not, write to the Free Software
## Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
## or see http://www.gnu.org/licenses/gpl.txt


"""
========
trefnydd
========

Package:   trefnydd
Module:    audioFileInterface
Copyright: 2007 Ivan A. Uemlianin
Contact:   ivan@llaisdy.com
Licence:   GNU GPL (see doc/gpl.txt)

audioFileInterface
==================

"""

import os, wave

from trefnydd.utils import location

class AudioFileInterface(location.Located):
    def __init__(self, debug=False):
        location.Located.__init__(self, debug)
        self._audioFn = ''
        self.audioFormat = ''
        self.audioData = None
        self.sampleRate = 0
        self.sampleSize = 0

    def setAudioFn(self, fn=''):
        if fn:
            fullfn = os.path.abspath(fn)
            if not self.homeDir or not fullfn.startswith(self.homeDir):
                self._audioFn = fn
            else:
                self._audioFn = fn[len(self.homeDir):]
                if self._audioFn and self._audioFn[0] == '/':
                    self._audioFn = self._audioFn[1:]
        else:
            self._audioFn = '%s.wav' % self.name
        self.changed = True

    def getAudioFn(self):
        if self.homeDir:
            fn = '%s.wav' % self.name
            thisAudioFn = os.path.join(self.homeDir, fn)
        elif self._audioFn:
            thisAudioFn =  os.path.join(self.homeDir, self._audioFn)
        else:
            self.setAudioFn()
            thisAudioFn = self.getAudioFn()
        return thisAudioFn

    audioFn = property(getAudioFn, setAudioFn)

    def read(self, fn):
        pass

    def write(self, fn='', start_t=0, stop_t=0):
        """\
        start_t, stop_t: seconds
        Writes data to file.
        """
        pass

    def getStartEnd(selfwindowSize=0.25):
        """\
        # todo: what does this do!?
        """
        pass

    def setStartEnd(self, newStart, newEnd):
        """\
        
        """
        start_b = int(newStart * self.sampleRate
                      * self.sampleSize)
        # the following checks start point is at a sample start point
        if math.modf(start_b/(self.sampleSize*1.0))[0] != 0.0:
            start_b += 1
        stop_b = int(newEnd * self.sampleRate
                     * self.sampleSize)
        # pad either end with artificial silence
        self.audioData = self.audioData[start_b:stop_b]

    def silence(self, t):
        return '\x00' * t * self.sampleSize * self.sampleRate

    def print_details(self, comment=''):
        if comment: comment = ' (%s)' % comment
        print 'audioFileInterface.print_details%s:\n\taudioFn: %s\n\tsampleRate: %s' \
              % (comment, self.audioFn, self.sampleRate)

    

class WavFileInterface(AudioFileInterface):
    
    def __init__(self, debug=False):
        AudioFileInterface.__init__(self, debug)
        self.audioFormat = 'wav'
    
    def read(self, fn):
        data = wave.open(fn)
        nframes = data.getnframes()
        self.audioData = data.readframes(nframes)
        self.sampleRate = data.getframerate()
        self.sampleSize = data.getsampwidth()
        self.audioFn = fn
        if self.debug is True: self.print_details('Reading')

    def write(self, fn='', start_t=0, stop_t=0):
        """\
        start_t, stop_t: seconds
        Writes data to file.
        """
        if not fn: fn = self.audioFn
        out = wave.open(fn, 'wb')
        out.setnchannels(1) # assume mono
        out.setframerate(self.sampleRate)
        out.setsampwidth(self.sampleSize)
        if not (start_t or stop_t): audioData = self.audioData
        else:
            start_b = int(start_t * self.sampleRate
                          * self.sampleSize)
            # check start point is at a sample start point
            if math.modf(start_b/(self.sampleSize*1.0))[0] != 0.0:
                start_b += 1
            stop_b = int(stop_t * self.sampleRate * self.sampleSize)
            # pad either end with artificial silence
            audioData = self.audioData[start_b:stop_b]
            if start_b != 0:
                audioData = self.silence(0.5) + audioData
            audioData += self.silence(0.5)
        out.writeframes(audioData)
        out.close()
        if self.debug is True: self.print_details('Writing')

    def getStartEnd(self, windowSize=0.25):
        noiseTimes = []
        data = wave.open(self.getAudioFn())
        width = data.getsampwidth()
        frate = data.getframerate()
        nframes = data.getnframes()
        endTime = nframes*1.0/frate
        rmsFull = audioop.rms(data.readframes(nframes), width)
        if self.debug: print 'rmsFull = ', rmsFull
        data.rewind()
        window = int(frate*windowSize)
        step = window/2
        while data.tell() < nframes-step:
            sample = data.readframes(window)
            rms = audioop.rms(sample, width)
            if rms > rmsFull/10:
                now = data.tell()*1.0/frate
                noiseTimes.append(now)
                if self.debug: print '%s\t%s' % (rms, now)
            elif self.debug: print rms
            data.setpos(data.tell()-step)
        return noiseTimes[0]-(windowSize/2), noiseTimes[-1], endTime


def main():
    pass

if __name__ == '__main__':
    main()
 
